@@ -1,3 +1,9 @@
+__all__ = ['string_converter', 'datetuple_converter', 'boolean_converter',
+           'file_converter','json_converter', 'converter']
+
+import csv
+from cStringIO import StringIO
+from datetime import date, datetime, time
 from simplegeneric import generic
 import schemaish
 
@@ -6,7 +12,8 @@ try:
     haveDecimal = True
 except ImportError:
     haveDecimal = False
-from datetime import date, time
+
+from convertish.util import SimpleTZInfo
 
 
 class ConvertError(Exception):
@@ -22,6 +29,11 @@ class ConvertError(Exception):
         return self.message
     __unicode__ = __str__
 
+    # Hide Python 2.6 deprecation warnings.
+    def _get_message(self): return self._message
+    def _set_message(self, message): self._message = message
+    message = property(_get_message, _set_message)
+
 
 class Converter(object):
     
@@ -33,13 +45,17 @@ class Converter(object):
         """
         convert from i.e. for NumberToString converter - from number to string
         """
+        raise NotImplementedError()
     
     def to_type(self, value, converter_options={}):
         """
         convert to i.e. for NumberToString converter - to number from string
         """
+        raise NotImplementedError()
+
 
 class NullConverter(Converter):
+
     def from_type(self, value, converter_options={}):
         return value
     
@@ -48,6 +64,7 @@ class NullConverter(Converter):
 
 
 class NumberToStringConverter(Converter):
+
     cast = None
     type_string = 'number'
     
@@ -84,19 +101,32 @@ if haveDecimal:
         cast = decimal.Decimal
 
 
-
 class FileToStringConverter(Converter):
+    """
+    Convert between a text File and a string.
+
+    The file's content is assumed to be a UTF-8 encoded string. Anything else
+    will almost certainly break the code and/or page.
+
+    Converting from a string to a File instance returns a new File with a
+    default name, content.txt, of type text/plain.
+    """
     
     def from_type(self, value, converter_options={}):
         if value is None:
             return None
-        return value.filename
+        if not value.file:
+            raise ValueError('Cannot convert to string without a file-like '
+                             'object to read from')
+        return value.file.read().decode('utf-8')
         
     def to_type(self, value, converter_options={}):
         if value is None:
             return None
         value = value.strip()
-        return schemaish.type.File(None,value,None)
+        return schemaish.type.File(StringIO(value.encode('utf-8')),
+                                   'content.txt', 'text/plain')
+
     
 class BooleanToStringConverter(Converter):
     
@@ -114,6 +144,7 @@ class BooleanToStringConverter(Converter):
         if value not in ('True', 'False'):
             raise ConvertError('%r should be either True or False'%value)
         return value == 'True'
+
     
 class DateToStringConverter(Converter):
     
@@ -126,20 +157,9 @@ class DateToStringConverter(Converter):
         if value is None:
             return None
         value = value.strip()
-        return self.parseDate(value)
+        return _parse_date(value)
+
         
-    def parseDate(self, value):
-        try:
-            y, m, d = [int(p) for p in value.split('-')]
-        except ValueError:
-            raise ConvertError('Invalid date')
-        try:
-            value = date(y, m, d)
-        except ValueError, e:
-            raise ConvertError('Invalid date: '+str(e))
-        return value
-
-
 class TimeToStringConverter(Converter):
     
     def from_type(self, value, converter_options={}):
@@ -151,37 +171,77 @@ class TimeToStringConverter(Converter):
         if value is None:
             return None
         value = value.strip()
-        return self.parseTime(value)
+        return _parse_time(value)
         
-    def parseTime(self, value):
-        
-        if '.' in value:
-            value, ms = value.split('.')
-        else:
-            ms = 0
-            
-        try:
-            parts = value.split(':')  
-            if len(parts)<2 or len(parts)>3:
-                raise ValueError()
-            if len(parts) == 2:
-                h, m = parts
-                s = 0
-            else:
-                h, m, s = parts
-            h, m, s, ms = int(h), int(m), int(s), int(ms)
-        except:
-            raise ConvertError('Invalid time')
-        
-        try:
-            value = time(h, m, s, ms)
-        except ValueError, e:
-            raise ConvertError('Invalid time: '+str(e))
-            
-        return value
- 
+
+class DateTimeToStringConverter(Converter):
+
+    def from_type(self, value, converter_options={}):
+        return value.isoformat()
+
+    def to_type(self, value, converter_options={}):
+        d, t = value.split('T')
+        d = _parse_date(d)
+        t = _parse_time(t)
+        return datetime(d.year, d.month, d.day, t.hour, t.minute, t.second,
+                        t.microsecond, t.tzinfo)
 
     
+def _parse_date(value):
+    try:
+        y, m, d = [int(p) for p in value.split('-')]
+    except ValueError:
+        raise ConvertError('Invalid date')
+    try:
+        value = date(y, m, d)
+    except ValueError, e:
+        raise ConvertError('Invalid date: '+str(e))
+    return value
+
+
+def _parse_time(value):
+    
+    # Parse the timezone offset
+    if '+' in value:
+        value, tz = value.split('+')
+        tzdir = 1
+    elif '-' in value:
+        value, tz = value.split('-')
+        tzdir = -1
+    else:
+        tz = None
+    if tz:
+        hours, minutes = tz.split(':')
+        tz = SimpleTZInfo(tzdir*((int(hours)*60) + int(minutes)))
+
+    # Parse milliseconds.
+    if '.' in value:
+        value, ms = value.split('.')
+    else:
+        ms = 0
+        
+    # Parse hours, minutes and seconds.
+    try:
+        parts = value.split(':')  
+        if len(parts)<2 or len(parts)>3:
+            raise ValueError()
+        if len(parts) == 2:
+            h, m = parts
+            s = 0
+        else:
+            h, m, s = parts
+        h, m, s, ms = int(h), int(m), int(s), int(ms)
+    except:
+        raise ConvertError('Invalid time')
+    
+    try:
+        value = time(h, m, s, ms, tz)
+    except ValueError, e:
+        raise ConvertError('Invalid time: '+str(e))
+        
+    return value
+
+
 class DateToDateTupleConverter(Converter):
     
     def from_type(self, value, converter_options={}):
@@ -201,39 +261,43 @@ class DateToDateTupleConverter(Converter):
         except (TypeError, ValueError), e:
             raise ConvertError('Invalid date: '+str(e))
         return value
-        
-
 
 
 def getDialect(delimiter=','):
-    import csv
     class Dialect(csv.excel):
         def __init__(self, *a, **k):
             self.delimiter = k.pop('delimiter',',')
             csv.excel.__init__(self,*a, **k)
     return Dialect(delimiter=delimiter)
 
+
 def convert_csvrow_to_list(row, delimiter=','):
-    import cStringIO as StringIO
-    import csv
-    dialect = getDialect(delimiter=delimiter)
-    sf = StringIO.StringIO()
-    csvReader = csv.reader(sf, dialect=dialect)
-    sf.write(row)
+    sf = StringIO()
+    sf.write(row.encode('utf-8'))
     sf.seek(0,0)
-    return csvReader.next()
+    reader = csv.reader(sf, dialect=getDialect(delimiter=delimiter))
+    return list(_decode_row(reader.next()))
+
     
 def convert_list_to_csvrow(l, delimiter=','):
-    import cStringIO as StringIO
-    import csv
-    dialect = getDialect(delimiter=delimiter)
-    sf = StringIO.StringIO()
-    writer = csv.writer(sf, dialect=dialect)
-    writer.writerow(l)
+    sf = StringIO()
+    writer = csv.writer(sf, dialect=getDialect(delimiter=delimiter))
+    writer.writerow(list(_encode_row(l)))
     sf.seek(0,0)
-    return sf.read().strip()
+    return sf.read().strip().decode('utf-8')
 
 
+def _encode_row(row, encoding='utf-8'):
+    for cell in row:
+        if cell is not None:
+            cell = cell.encode(encoding)
+        yield cell
+
+
+def _decode_row(row, encoding='utf-8'):
+    for cell in row:
+        yield cell.decode(encoding)
+
         
 class SequenceToStringConverter(Converter):
     """
@@ -273,7 +337,6 @@ class SequenceToStringConverter(Converter):
             value =  [string_converter(self.schema_type.attr).from_type(v) \
                       for v in value]
             return convert_list_to_csvrow(value, delimiter=delimiter)
-        
     
     def to_type(self, value, converter_options={}):
         if value is None:
@@ -310,11 +373,14 @@ class SequenceToStringConverter(Converter):
                     for v in out]
 
 
-
 class TupleToStringConverter(Converter):
     """
-    I'd really like to have the converter options on the init but ruledispatch
+    Convert a tuple to and from a string.
+
+    XXX tim: I'd really like to have the converter options on the init but ruledispatch
     won't let me pass keyword arguments
+    XXX matt: the default to_type items should be configurable but None is
+    better than '' because it doesn't crash the item's converter ;-).
     """
     
     def __init__(self, schema_type, **k):
@@ -326,10 +392,7 @@ class TupleToStringConverter(Converter):
         delimiter = converter_options.get('delimiter',',')
         lineitems =  [string_converter(self.schema_type.attrs[n]).from_type(item) \
                       for n,item in enumerate(value)]
-        linestring = convert_list_to_csvrow(lineitems, delimiter=delimiter)
-
-        return linestring
-        
+        return convert_list_to_csvrow(lineitems, delimiter=delimiter)
     
     def to_type(self, value, converter_options={}):
         if value is None:
@@ -337,14 +400,21 @@ class TupleToStringConverter(Converter):
         value = value.strip()
         delimiter = converter_options.get('delimiter',',')
         l = convert_csvrow_to_list(value, delimiter=delimiter)
-        convl = [string_converter(self.schema_type.attrs[n]).to_type(v) \
-                 for n,v in enumerate(l[:len(self.schema_type.attrs)])]
-        return tuple(convl)
+        if len(l) > len(self.schema_type.attrs):
+            raise ConvertError('Too many arguments')
+        if len(l) < len(self.schema_type.attrs):
+            raise ConvertError('Too few arguments')
+        def convert_or_none(n, v):
+            v = v.strip()
+            if not v:
+                return None
+            return string_converter(self.schema_type.attrs[n]).to_type(v)
+        return tuple(convert_or_none(n, v) for (n, v) in enumerate(l))
 
 
 class TupleToListConverter(Converter):
 
-    def __init__(self, schem_type, **k):
+    def __init__(self, schema_type, **k):
         Converter.__init__(self, schema_type, **k)
 
     def from_type(self, value, converter_options={}):
@@ -357,19 +427,29 @@ class TupleToListConverter(Converter):
             return None
         return tuple(value)
 
+def converter(type):
+    if type == 'string':
+        return string_converter
+    if type == 'json':
+        return json_converter
+    if type == 'datetuple':
+        return datetuple_converter
+    if type == 'boolean':
+        return booleank_converter
+    if type == 'file':
+        return file_converter
+    else:
+        raise ValueError('no type converter for %s'%type)
 
 
 ####
 #
 #  String Converter
-
-    
     
 @generic
 def string_converter(schema_type):
     pass
 
-
 @string_converter.when_type(schemaish.String)
 def string_to_string(schema_type):
     return NullConverter(schema_type)
@@ -394,6 +474,10 @@ def date_to_string(schema_type):
 def time_to_string(schema_type):
     return TimeToStringConverter(schema_type)
 
+@string_converter.when_type(schemaish.DateTime)
+def datetime_to_string(schema_type):
+    return DateTimeToStringConverter(schema_type)
+
 @string_converter.when_type(schemaish.Sequence)
 def sequence_to_string(schema_type):
     return SequenceToStringConverter(schema_type)
@@ -410,11 +494,11 @@ def boolean_to_string(schema_type):
 def file_to_string(schema_type):
     return FileToStringConverter(schema_type)
 
+
 ####
 #
 #  Date Tuple Converter
 
-
 @generic
 def datetuple_converter(schema_type):
     pass
@@ -423,11 +507,11 @@ def datetuple_converter(schema_type):
 def date_to_datetuple(schema_type):
     return DateToDateTupleConverter(schema_type)
 
+
 ####
 #
 #  Boolean Converter
 
-
 @generic
 def boolean_converter(schema_type):
     pass
@@ -436,8 +520,6 @@ def boolean_converter(schema_type):
 def boolean_to_boolean(schema_type):
     return NullConverter(schema_type)
 
-
-
 @generic
 def file_converter(schema_type):
     pass
@@ -472,6 +554,7 @@ class DateToJSONConverter(Converter):
             raise ConvertError('Invalid date: '+str(e))
         return value
 
+
 class TimeToJSONConverter(Converter):
     
     def from_type(self, value, converter_options={}):
@@ -538,10 +621,3 @@ def boolean_to_json(schema_type):
 def file_to_json(schema_type):
     return FileToStringConverter(schema_type)
 
-
-
-
-__all__ = [
-    'string_converter', 'datetuple_converter',
-    'boolean_converter', 'file_converter','json_converter'
-    ]
